import { readFileSync } from 'fs';
import { dirname, extname } from 'path';
import { createCustomDebug } from '@linaria/logger';
import { buildOptions, getFileIdx, loadBabelOptions } from '@linaria/utils';
import withLinariaMetadata from '../utils/withLinariaMetadata';
import cachedParseSync from './helpers/cachedParseSync';
import loadLinariaOptions from './helpers/loadLinariaOptions';

const isModuleResolver = i => typeof i === 'object' && i !== null && i.key === 'module-resolver';

function runPreevalStage(babel, filename, code, options, perFileBabelConfig) {
  const pluginOptions = loadLinariaOptions(options.pluginOptions);
  const parseConfig = buildOptions(pluginOptions?.babelOptions, perFileBabelConfig);
  const fullParserOptions = loadBabelOptions(babel, filename, parseConfig);
  const file = cachedParseSync(babel, code, fullParserOptions);
  const transformPlugins = [[require.resolve('../plugins/preeval'), pluginOptions]];
  const moduleResolverPlugin = fullParserOptions.plugins?.find(isModuleResolver);

  if (moduleResolverPlugin) {
    transformPlugins.unshift(moduleResolverPlugin);
  }

  const transformConfig = buildOptions({
    envName: 'linaria',
    plugins: transformPlugins,
    sourceMaps: true,
    sourceFileName: filename,
    inputSourceMap: options.inputSourceMap,
    root: options.root,
    ast: true,
    babelrc: false,
    configFile: false
  });
  const result = babel.transformFromAstSync(file, code, { ...transformConfig,
    filename
  });

  if (!result || !result.ast?.program) {
    throw new Error('Babel transform failed');
  }

  return result;
}

function getMatchedRule(rules, filename, code) {
  for (let i = rules.length - 1; i >= 0; i--) {
    const rule = rules[i];

    if (!rule.test) {
      return rule;
    }

    if (typeof rule.test === 'function' && rule.test(filename, code)) {
      return rule;
    }

    if (rule.test instanceof RegExp && rule.test.test(filename)) {
      return rule;
    }
  }

  return {
    action: 'ignore'
  };
}

function prepareCode(babel, filename, originalCode, only, options, fileCache) {
  const log = createCustomDebug('transform', getFileIdx(filename));
  const pluginOptions = loadLinariaOptions(options.pluginOptions);
  const {
    action,
    babelOptions
  } = getMatchedRule(pluginOptions.rules, filename, originalCode);

  if (action === 'ignore') {
    log('stage-1:ignore', '');
    fileCache.set('*', {
      code: originalCode
    });
    return [originalCode, null];
  }

  const preevalStageResult = runPreevalStage(babel, filename, originalCode, options, babelOptions);

  if (only.length === 1 && only[0] === '__linariaPreval' && !withLinariaMetadata(preevalStageResult.metadata)) {
    log('stage-1:evaluator:end', 'no metadata');
    return [preevalStageResult.code, null, preevalStageResult.metadata];
  }

  log('stage-1:preeval', 'metadata %O', preevalStageResult.metadata); // Action can be a function or a module name

  const evaluator = typeof action === 'function' ? action : require(require.resolve(action, {
    paths: [dirname(filename)]
  })).default;
  log('stage-1:evaluator:start', 'using %s', evaluator.name);
  const result = evaluator(filename, pluginOptions, preevalStageResult.code, only, babel);
  log('stage-1:evaluator:end', '');
  return [...result, preevalStageResult.metadata];
}

function processQueueItem(babel, item, codeCache, options) {
  if (!item) {
    return undefined;
  }

  const pluginOptions = loadLinariaOptions(options.pluginOptions);
  const results = new Set();
  const {
    name,
    only,
    code
  } = item;

  if (!codeCache.has(name)) {
    codeCache.set(name, new Map());
  }

  const log = createCustomDebug('transform', getFileIdx(name));
  const extension = extname(name);

  if (!pluginOptions.extensions.includes(extension)) {
    log('init', `${name} is ignored. If you want it to be processed, you should add '${extension}' to the "extensions" option.`);
    return undefined;
  }

  log('init', `${name} (${only.join(', ')})\n${code}`);
  const uncachedExports = new Set(only);
  const fileCache = codeCache.get(name);
  uncachedExports.forEach(token => {
    if (fileCache.has(token)) {
      uncachedExports.delete(token);
      results.add(fileCache.get(token));
    }
  });

  if (uncachedExports.size === 0) {
    // Already processed
    return {
      imports: null,
      name,
      results: Array.from(results)
    };
  }

  const remainExports = Array.from(uncachedExports);
  log('stage-1', `>> (${remainExports.join(', ')})`);
  const [preparedCode, imports, metadata] = prepareCode(babel, name, code, remainExports, options, fileCache);

  if (code === preparedCode) {
    log('stage-1', `<< (${remainExports.join(', ')})\n === no changes ===`);
  } else {
    log('stage-1', `<< (${remainExports.join(', ')})\n${preparedCode}`);
  }

  const result = {
    metadata,
    code: preparedCode
  };
  results.add(result);
  remainExports.forEach(token => {
    fileCache.set(token, result);
  });
  if (preparedCode === '') return undefined;
  return {
    imports,
    name,
    results: Array.from(results)
  };
}

export function prepareForEvalSync(babel, resolveCache, codeCache, resolve, resolvedFile, options, stack = []) {
  const processed = processQueueItem(babel, resolvedFile, codeCache, options);
  if (!processed) return undefined;
  const {
    imports,
    name,
    results
  } = processed;
  const log = createCustomDebug('transform', getFileIdx(name));
  const queue = [];
  imports?.forEach((importsOnly, importedFile) => {
    try {
      const resolved = resolve(importedFile, name, stack);
      log('stage-1:sync-resolve', `✅ ${importedFile} -> ${resolved}`);
      resolveCache.set(`${name} -> ${importedFile}`, `${resolved}\0${importsOnly.join(',')}`);
      const fileContent = readFileSync(resolved, 'utf8');
      queue.push({
        name: resolved,
        only: importsOnly,
        code: fileContent
      });
    } catch (err) {
      log('stage-1:sync-resolve', `❌ cannot resolve ${importedFile}: %O`, err);
    }
  });
  queue.forEach(item => {
    prepareForEvalSync(babel, resolveCache, codeCache, resolve, item, options, [name, ...stack]);
  });
  return Array.from(results);
}
const mutexes = new Map();
/**
 * Parses the specified file and recursively all its dependencies,
 * finds tags, applies eval-time replacements, removes dead code.
 */

export default async function prepareForEval(babel, resolveCache, codeCache, resolve, file, options, stack = []) {
  const resolvedFile = await file;
  const mutex = resolvedFile ? mutexes.get(resolvedFile.name) : null;

  if (mutex) {
    await mutex;
  }

  const processed = processQueueItem(babel, resolvedFile, codeCache, options);
  if (!processed) return undefined;
  const {
    imports,
    name,
    results
  } = processed;
  const log = createCustomDebug('transform', getFileIdx(name));
  const promises = [];
  imports?.forEach((importsOnly, importedFile) => {
    const promise = resolve(importedFile, name, stack).then(resolved => {
      if (resolved === null) {
        log('stage-1:resolve', `✅ ${importedFile} is ignored`);
        return null;
      }

      log('stage-1:async-resolve', `✅ ${importedFile} -> ${resolved}`);
      const resolveCacheKey = `${name} -> ${importedFile}`;
      const cached = resolveCache.get(resolveCacheKey);
      const importsOnlySet = new Set(importsOnly);

      if (cached) {
        const [, cachedOnly] = cached.split('\0');
        cachedOnly?.split(',').forEach(token => {
          importsOnlySet.add(token);
        });
      }

      resolveCache.set(resolveCacheKey, `${resolved}\0${[...importsOnlySet].join(',')}`);
      const fileContent = readFileSync(resolved, 'utf8');
      return {
        name: resolved,
        only: importsOnly,
        code: fileContent
      };
    }, err => {
      log('stage-1:async-resolve', `❌ cannot resolve ${importedFile}: %O`, err);
      return null;
    });
    promises.push(prepareForEval(babel, resolveCache, codeCache, resolve, promise, options, [name, ...stack]));
  });
  const promise = Promise.all(promises).then(() => {});
  mutexes.set(name, promise);
  await promise;
  mutexes.delete(name);
  return Array.from(results);
}
//# sourceMappingURL=1-prepare-for-eval.js.map